"""
Copyright (c) 2024 valmi.io <https://github.com/valmi-io>

Created Date: Friday, March 10th 2023, 8:49:37 pm
Author: Rajashekar Varkala @ valmi.io

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import os
import sys
from os.path import join
import subprocess
import io
from typing import Any, Dict
import uuid
from pydantic import UUID4
import requests
from requests.adapters import HTTPAdapter, Retry

from valmi_connector_lib.common.logs import SingletonLogWriter, TimeAndChunkEndFlushPolicy
from valmi_connector_lib.common.samples import SampleWriter
from valmi_connector_lib.valmi_protocol import ValmiFinalisedRecordMessage

# TODO: Constants - need to become env vars
MAGIC_NUM = 0x7FFFFFFF
HTTP_TIMEOUT = 3  # seconds
MAX_HTTP_RETRIES = 5
CONNECTOR_STRING = "src"

state_file_path = None
loaded_state = None


# desanitise uuid
def du(uuid_str: str) -> UUID4:
    return uuid.UUID(uuid_str.replace("_", "-"))


class ConnectorState:
    def __init__(self, run_time_args={}) -> None:
        self.num_chunks = 1
        self.records_in_chunk = 0
        self.run_time_args = run_time_args
        self.total_records = 0
        
    def reset_chunk_id_from_state(self, state):
        if state is not None \
            and 'state' in state \
                and 'data' in state['state'] \
                    and 'chunk_id' in state['state']['data']:
            self.num_chunks = state['state']['data']['chunk_id'] + 1
        else:
            self.num_chunks = 1
        self.total_records = (self.num_chunks - 1) * self.run_time_args["chunk_size"]

    def register_chunk(self):
        self.num_chunks = self.num_chunks + 1
        self.records_in_chunk = 0

    def register_record(self):
        self.records_in_chunk = self.records_in_chunk + 1
        self.total_records = self.total_records + 1


class NullEngine:
    def __init__(self) -> None:
        self.connector_state = None
        pass

    def error(self, msg="error"):
        pass

    def success(self):
        pass

    def metric(self):
        pass

    def metric_ext(self, metric_json, commit=False):
        pass

    def current_run_details(self):
        return {}

    def abort_required(self):
        return False


class Engine(NullEngine):
    def __init__(self, *args, **kwargs):
        super(Engine, self).__init__(*args, **kwargs)
        self.engine_url = os.environ["ACTIVATION_ENGINE_URL"]

        self.session_with_retries = requests.Session()
        # retry on all errors except 400 and 401
        status_forcelist = tuple(x for x in requests.status_codes._codes if x > 400 and x not in [400, 401])
        retries = Retry(total=MAX_HTTP_RETRIES, backoff_factor=5, status_forcelist=status_forcelist)
        self.session_with_retries.mount("http://", HTTPAdapter(max_retries=retries))
        self.session_with_retries.mount("https://", HTTPAdapter(max_retries=retries))

        self.session_without_retries = requests.Session()
        status_forcelist = []
        retries = Retry(total=0, backoff_factor=5, status_forcelist=status_forcelist)
        self.session_without_retries.mount("http://", HTTPAdapter(max_retries=retries))
        self.session_without_retries.mount("https://", HTTPAdapter(max_retries=retries))

        run_time_args = self.current_run_details()
        self.connector_state = ConnectorState(run_time_args=run_time_args)

    def current_run_details(self):
        sync_id = du(os.environ.get("DAGSTER_RUN_JOB_NAME", "cf280e5c-1184-4052-b089-f9f41b25138e"))
        #sync_id = du(os.environ.get("DAGSTER_RUN_JOB_NAME", "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"))
        r = self.session_with_retries.get(
            f"{self.engine_url}/syncs/{sync_id}/runs/current_run_details/{CONNECTOR_STRING}",
            timeout=HTTP_TIMEOUT,
        )
        return r.json()

    def metric(self, commit=False):
        self.metric_ext({"succeeded": self.connector_state.records_in_chunk}, commit=commit)

    def metric_ext(self, metric_json, commit=False):
        print("Sending metric")
        payload = {
            "sync_id": self.connector_state.run_time_args["sync_id"],
            "run_id": self.connector_state.run_time_args["run_id"],
            "chunk_id": self.connector_state.num_chunks,
            "connector_id": CONNECTOR_STRING,
            "metrics": metric_json,
            "commit": commit,
        }

        print("payload ", payload)
        if commit:
            r = self.session_with_retries.post(
                f"{self.engine_url}/metrics/",
                timeout=HTTP_TIMEOUT,
                json=payload,
            )
            r.raise_for_status()
        else:
            try:
                r = self.session_without_retries.post(
                    f"{self.engine_url}/metrics/",
                    timeout=HTTP_TIMEOUT,
                    json=payload,
                )
            except Exception:
                pass

    def error(self, msg="error"):
        print("sending error ", msg)
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/status/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json={"status": "failed", "message": msg},
        )
        r.raise_for_status()

    def success(self):
        print("sending success ")
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/status/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json={"status": "success"},
        )
        r.raise_for_status()

    def abort_required(self):
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.get(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/synchronize_connector_engine", timeout=HTTP_TIMEOUT
        )
        return r.json()["abort_required"]

    def checkpoint(self, state):
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/state/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json=state,
        )
        r.raise_for_status()


class NullWriter:
    def __init__(self, engine: NullEngine) -> None:
        pass

    def write(self, record, last=False):
        pass

    def flush(self, last=False):
        pass

    def finalize(self):
        pass


class StdoutWriter(NullWriter):
    pass


class StoreWriter(NullWriter):
    def __init__(self, engine: NullEngine) -> None:
        self.engine = engine
        self.connector_state: ConnectorState = self.engine.connector_state
        store_config = json.loads(os.environ["VALMI_INTERMEDIATE_STORE"])
        if store_config["provider"] == "local":
            path_name = join(store_config["local"]["directory"], self.connector_state.run_time_args["run_id"], "data")
            os.makedirs(path_name, exist_ok=True)

            self.path_name = path_name
            self.records = []

    def write(self, record, last=False):
        self.records.append(record)
        self.connector_state.register_record()
        if self.connector_state.records_in_chunk >= self.connector_state.run_time_args["chunk_size"]:
            self.flush(last=False)
            self.records = []
            self.engine.metric(commit=True)
            self.connector_state.register_chunk()
        elif self.connector_state.records_in_chunk % self.connector_state.run_time_args["records_per_metric"] == 0:
            self.engine.metric(commit=False)

    def flush(self, last=False):
        # list_dir = sorted([f.lower() for f in os.listdir(self.path_name)], key=lambda x: int(x[:-5]))
        new_file_name = f"{MAGIC_NUM}.vald" if last else f"{self.engine.connector_state.num_chunks}.vald"
        with open(join(self.path_name, new_file_name), "w") as f:
            for record in self.records:
                f.write(json.dumps(record))
                f.write("\n")

    def finalize(self):
        self.flush(last=True)
        self.engine.metric(commit=True)


class DefaultHandler:
    def __init__(
        self, engine: Engine = None, store_writer: StoreWriter = None, stdout_writer: StdoutWriter = None
    ) -> None:
        self.engine = engine
        self.store_writer = store_writer
        self.stdout_writer = stdout_writer
        pass

    def handle(self, record):
        print(json.dumps(record))
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()

    def finalize(self):
        pass


class LogHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(LogHandler, self).__init__(*args, **kwargs)

    def handle(self, record):
        log_str = json.dumps(record)
        print(log_str)
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().write(log_str)


class CheckpointHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(CheckpointHandler, self).__init__(*args, **kwargs)

    def handle(self, record):
        print(json.dumps(record))
        
        if os.environ.get('MODE', 'any') == 'etl':
            record['state']['data']['chunk_id']= self.engine.connector_state.num_chunks - 1 # Not oprating on chunk boundary -- fix
            self.store_writer.write(record)

        self.engine.checkpoint(record)
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().data_chunk_flush_callback()
        SampleWriter.data_chunk_flush_callback()


class RecordHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(RecordHandler, self).__init__(*args, **kwargs)

    def handle(self, record):

        # HACK:
        if not isinstance(record, ValmiFinalisedRecordMessage):
            record["record"]["rejected"] = False
            record["record"]["metric_type"] = "succeeded"
        # 

        if not record["record"]["rejected"]:
            self.store_writer.write(record)

        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()
        # print(record)
        sample_writer = SampleWriter.get_writer_by_metric_type(metric_type=record["record"]["metric_type"])
        sample_writer.write(record)

    def finalize(self):
        self.store_writer.finalize()


class TraceHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(TraceHandler, self).__init__(*args, **kwargs)

    def handle(self, record):
        print(json.dumps(record))
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()
        if record["trace"]["type"] == "ERROR":
            self.engine.error(record["trace"]["error"]["message"])
            sys.exit(0)
        elif record["trace"]["type"] == "ESTIMATE":
            self.engine.metric_ext(
                {record["trace"]["estimate"]["row_kind"]: record["trace"]["estimate"]["row_estimate"]}, commit=True
            )


handlers = {
    "LOG": LogHandler,
    "STATE": CheckpointHandler,
    "RECORD": RecordHandler,
    "default": DefaultHandler,
    "TRACE": TraceHandler,
}


def get_airbyte_command():
    entrypoint_str = os.environ["VALMI_ENTRYPOINT"]
    entrypoint = entrypoint_str.split(" ")

    airbyte_command = sys.argv[3]
    for i, arg in enumerate(sys.argv[1:]):
        if i >= len(entrypoint):
            airbyte_command = arg
            break

    return airbyte_command


def get_config_file_path():
    # TODO: do this better
    config_file_path = None
    for i, arg in enumerate(sys.argv):
        if arg == "--config":
            config_file_path = sys.argv[i + 1]
            break

    return config_file_path


def populate_run_time_args(airbyte_command, engine, config_file_path):
    if airbyte_command == "read":
        run_time_args = engine.current_run_details()

        with open(config_file_path, "r") as f:
            config = json.loads(f.read())

        with open(config_file_path, "w") as f:
            config["run_time_args"] = run_time_args
            f.write(json.dumps(config))

        print("checking state")
        if 'state' in run_time_args:
            # create a new state file alongside the config file
            state_file_path = os.path.join(os.path.dirname(config_file_path), 'state.json')
            set_state_file_path(state_file_path)
            set_loaded_state(run_time_args['state'])
            print(loaded_state)
            engine.connector_state.reset_chunk_id_from_state(loaded_state)
            print("num_chunks", engine.connector_state.num_chunks)
            with open(state_file_path, "w") as f:
                f.write(json.dumps(run_time_args['state']))


def sync_engine_for_error(proc: subprocess, engine: NullEngine):
    if engine.abort_required():
        proc.kill()
        sys.exit(0)  # Not this connector's fault
        

def set_state_file_path(file_path: str):
    global state_file_path
    state_file_path = file_path


def set_loaded_state(state: Dict[str, Any]):
    global loaded_state
    loaded_state = state


def is_state_available():
    global state_file_path
    return state_file_path is not None


def main():
    airbyte_command = get_airbyte_command()
    config_file = get_config_file_path()

    if airbyte_command is None or (airbyte_command != "spec" and config_file is None):
        sys.exit(5)

    # if arg in read, write:
    # read checkpoint from the engine

    engine = None
    if airbyte_command == "read":
        engine = Engine()
        store_writer = StoreWriter(engine)
    else:
        engine = NullEngine()
        store_writer = NullWriter(engine)

    # populate run_time_args
    populate_run_time_args(airbyte_command, engine, config_file_path=config_file)

    

    if airbyte_command == "read":
        # initialize LogWriter
        SingletonLogWriter(os.environ["VALMI_INTERMEDIATE_STORE"],
                           TimeAndChunkEndFlushPolicy(os.environ["VALMI_INTERMEDIATE_STORE"]),
                           engine.connector_state.run_time_args["sync_id"],
                           engine.connector_state.run_time_args["run_id"],
                           CONNECTOR_STRING)

        # initialize SampleWriter
        SampleWriter.get_writer_by_metric_type(store_config_str=os.environ["VALMI_INTERMEDIATE_STORE"],
                                               sync_id=engine.connector_state.run_time_args["sync_id"],
                                               run_id=engine.connector_state.run_time_args["run_id"],
                                               connector=CONNECTOR_STRING)

    stdout_writer = StdoutWriter(engine)

    # initialize handlers
    for key in handlers.keys():
        handlers[key] = handlers[key](engine=engine, store_writer=store_writer, stdout_writer=stdout_writer)

    # create the subprocess
    subprocess_args = sys.argv[1:]
    if is_state_available():
        subprocess_args.append("--state")
        subprocess_args.append(state_file_path)
    proc = subprocess.Popen(
        subprocess_args,
        stdout=subprocess.PIPE,
    )

    # check engine errors every CHUNK_SIZE records
    

    record_types = handlers.keys()
    for line in io.TextIOWrapper(proc.stdout, encoding="utf-8"):  # or another encoding
        if line.strip() == "":
            continue
        
        
        json_record = json.loads(line)
        if json_record["type"] not in record_types:
            handlers["default"].handle(json_record)
        else:
            handlers[json_record["type"]].handle(json_record)

        # sync with engine for any errors - in the future -> to apply backpressure to the connector
        if airbyte_command == "read":
            if engine.connector_state.records_in_chunk % engine.connector_state.run_time_args["chunk_size"] == 0:
                sync_engine_for_error(proc, engine=engine)

    return_code = proc.poll()

    # littered code with log flushes. What's better?
    if SingletonLogWriter.instance() is not None:
        SingletonLogWriter.instance().check_for_flush()

    if return_code is not None and return_code != 0:
        engine.error("Connector exited with non-zero return code")
        sys.exit(return_code)
    else:
        if airbyte_command == "read":
            store_writer.finalize()
        engine.success()


if __name__ == "__main__":
    main()



def get_metric_type(sync_op: str):
    """
    Returns the metric type for the given sync operation.
    """
    if sync_op.endswith("e"):
        return f"{sync_op}d"
    elif (sync_op.endswith("ed")):
        return f"{sync_op}"
    else:
        return f"{sync_op}d"



"""
Copyright (c) 2024 valmi.io <https://github.com/valmi-io>

Created Date: Monday, March 13th 2023, 9:59:01 am
Author: Rajashekar Varkala @ valmi.io

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import os
from os.path import join
import time

from valmi_connector_lib.common.logs import SingletonLogWriter
from valmi_connector_lib.common.samples import SampleWriter

from .engine import NullEngine, ConnectorState, Engine

# TODO: Constants - need to become env vars
MAGIC_NUM = 0x7FFFFFFF


class StoreWriter:
    def __init__(self) -> None:
        pass

    def store(self, record):
        print(record)


class StdoutWriter:
    def __init__(self) -> None:
        pass

    def store(self, record):
        print(record)


class StoreReader:
    def __init__(self, engine: NullEngine, state: str) -> None:
        self.engine = engine
        self.connector_state: ConnectorState = self.engine.connector_state
        self.loaded_state = state

        store_config = json.loads(os.environ["VALMI_INTERMEDIATE_STORE"])
        if store_config["provider"] == "local":
            path_name = join(store_config["local"]["directory"], self.connector_state.run_time_args["run_id"], "data")
            os.makedirs(path_name, exist_ok=True)
            self.path_name = path_name
            self.last_handled_fn = self.get_file_name_from_chunk_id(self.read_chunk_id_checkpoint())

    def read(self):
        while True:
            if not os.path.exists(self.path_name):
                time.sleep(1)
                yield ""
            list_dir = sorted([f.lower() for f in os.listdir(self.path_name)], key=lambda x: int(x[:-5]))
            for fn in list_dir:
                if self.last_handled_fn is not None and int(fn[:-5]) <= int(self.last_handled_fn[:-5]):
                    continue
                if fn.endswith(".vald"):
                    with open(join(self.path_name, fn), "r") as f:
                        for line in f.readlines():
                            # print("yiedling", line)
                            yield line

                    self.last_handled_fn = fn
                # print(fn)
                # print(f"magic num {MAGIC_NUM}")
                if fn.startswith(f"{MAGIC_NUM}"):
                    # print("returning")
                    return

                # Check for abort condition after reading a file
                if self.engine.abort_required():
                    return
                
            # Check for abort condition after exhausting files in the folder
            if self.engine.abort_required():
                return
            time.sleep(3)
            yield ""

    def read_chunk_id_checkpoint(self):
        # TODO: connector_state is not being used for destination, clean it up.
        if self.loaded_state is not None \
                and 'state' in self.loaded_state \
                and 'data' in self.loaded_state['state'] \
                and 'chunk_id' in self.loaded_state['state']['data']:
            return self.loaded_state['state']['data']['chunk_id']
        return None

    def get_file_name_from_chunk_id(self, chunk_id):
        if chunk_id is not None:
            return f"{chunk_id}.vald"
        return None


class DefaultHandler:
    def __init__(
        self,
        engine: Engine = None,
        store_writer: StoreWriter = None,
        stdout_writer: StdoutWriter = None,
        store_reader: StoreReader = None,
    ) -> None:
        self.engine = engine
        self.store_writer = store_writer
        self.stdout_writer = stdout_writer
        self.store_reader = store_reader
        pass

    def handle(self, record) -> bool:
        print(json.dumps(record))
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()
        return True


class LogHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(LogHandler, self).__init__(*args, **kwargs)

    def handle(self, record) -> bool:
        log_str = json.dumps(record)
        print(log_str)
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().write(log_str)
        return


class CheckpointHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(CheckpointHandler, self).__init__(*args, **kwargs)

    def handle(self, record) -> bool:
        print("Checkpoint seen")
        print(record)

        if os.environ.get('MODE', 'any') == 'etl':
            return True
        else :
            records_delivered = record["state"]["data"]["records_delivered"]
            finished = record["state"]["data"]["finished"]
            commit_state = record["state"]["data"]["commit_state"]
            commit_metric = record["state"]["data"]["commit_metric"]

            total_records = 0
            for k, v in records_delivered.items():
                total_records += v

            self.engine.connector_state.register_records(total_records)

            if commit_metric:
                self.engine.metric_ext(records_delivered, record["state"]["data"]["chunk_id"], commit=True)
                # self.engine.connector_state.register_chunk()
            if commit_state:
                self.engine.checkpoint(record)
                if SingletonLogWriter.instance() is not None:
                    SingletonLogWriter.instance().data_chunk_flush_callback()
                SampleWriter.data_chunk_flush_callback()
            else:
                if SingletonLogWriter.instance() is not None:
                    SingletonLogWriter.instance().check_for_flush()

            return True


class RecordHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(RecordHandler, self).__init__(*args, **kwargs)

    def handle(self, record) -> bool:
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()

        sample_writer = SampleWriter.get_writer_by_metric_type(metric_type=record["record"]["metric_type"])
        sample_writer.write(record)
        return True  # to continue reading


class TraceHandler(DefaultHandler):
    def __init__(self, *args, **kwargs):
        super(TraceHandler, self).__init__(*args, **kwargs)

    def handle(self, record):
        print(json.dumps(record))
        if SingletonLogWriter.instance() is not None:
            SingletonLogWriter.instance().check_for_flush()
        self.engine.error(record["trace"]["error"]["message"])
        return False



'''
Copyright (c) 2024 valmi.io <https://github.com/valmi-io>

Created Date: Monday, August 21st 2023, 3:23:12 pm
Author: Rajashekar Varkala @ valmi.io

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

from abc import abstractmethod
import json
from os.path import join
import time
import os
from .constants import MAGIC_DELIM


class LogWriter:

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, store_config_str, flush_policy, sync_id, run_id, connector):
        self.flush_policy = flush_policy
        self.store_config = json.loads(store_config_str)
        self.lines = []
        self.sync_id = sync_id
        self.run_id = run_id
        self.connector = connector

    def reset(self):
        self.lines.clear()
        self.flush_policy.reset()

    def check_for_flush(self):
        if self.flush_policy.roll_over_file_required():
            self.flush()
            self.reset()

        elif self.flush_policy.flush_required():
            self.flush()

    def write(self, log_record):
        current_time = self.flush_policy.newline(log_record)
        self.lines.append((current_time, log_record))

        self.check_for_flush()

    def flush(self):
        if len(self.lines) > 0:
            # TODO: do cloud push for aws, s3 when we do k8s
            file_path = join(self.store_config['local']['directory'],
                             self.run_id, "logs", self.connector, self.flush_policy.get_current_file_name())
            dir_name = os.path.dirname(file_path)
            os.makedirs(dir_name, exist_ok=True)
            with open(file_path, "w") as f:
                for line in self.lines:
                    f.write(f"{line[0]}")
                    f.write(MAGIC_DELIM)
                    f.write(line[1])
                    f.write("\n")

    def data_chunk_flush_callback(self):
        self.flush_policy.set_data_chunk_flushed()
        self.check_for_flush()


class FlushPolicy:
    def __init__(self, store_config_str: str):
        self.store_config = json.loads(store_config_str)

    @abstractmethod
    def flush_required(self):
        pass

    @abstractmethod
    def newline(self, log_record):
        pass

    @abstractmethod
    def roll_over_file_required(self):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def get_current_file_name(self):
        pass


class TimeAndChunkEndFlushPolicy(FlushPolicy):
    def __init__(self, *args, **kwargs):
        super(TimeAndChunkEndFlushPolicy, self).__init__(*args, **kwargs)
        self.reset()

    def flush_required(self):
        if (self.first_line_time is not None
                and self.timestamp_changed
                and self.last_line_time - self.first_line_time > self.store_config['local']['log_flush_interval'] * 1000000) \
                or self.data_chunk_flushed:
            return True
        return False
    
    def roll_over_file_required(self):
        if (self.num_lines >= self.store_config['local']['max_lines_per_file_hint']):
            return True
        return False

    def newline(self, log_record):
        self.num_lines += 1
        current_time = int(time.time() * 1000000)  # microseconds
        if self.first_line_time is None:
            self.first_line_time = current_time

        if self.last_line_time is not None and current_time != self.last_line_time:
            self.timestamp_changed = True
        else:
            self.timestamp_changed = False

        self.last_line_time = current_time
        return current_time

    def set_data_chunk_flushed(self):
        self.data_chunk_flushed = True

    def reset(self):
        self.first_line_time = None
        self.last_line_time = None
        self.timestamp_changed = False
        self.data_chunk_flushed = False
        self.num_lines = 0

    def get_current_file_name(self):
        return f"{self.first_line_time}.vall"


class SingletonLogWriter(LogWriter):
    __initialized = False

    def __new__(cls, *args, **kwargs) -> object:
        if not hasattr(cls, "_instance"):
            cls._instance = super(SingletonLogWriter, cls).__new__(cls, args, kwargs)
        return cls._instance

    def __init__(self, store_config, flush_policy, sync_id, run_id, connector):
        if SingletonLogWriter.__initialized:
            return

        SingletonLogWriter.__initialized = True
        super(SingletonLogWriter, self).__init__(store_config, flush_policy, sync_id, run_id, connector)

    @classmethod
    def instance(cls):
        if hasattr(cls, "_instance"):
            return cls._instance
        return None


def main():
    store_config_str = "{\"provider\": \"local\", \"local\": {\"directory\": \"/tmp/shared_dir/test_logs\", \"max_lines_per_file_hint\" : 100, \"log_flush_interval\": 5}}"
    SingletonLogWriter(store_config_str, TimeAndChunkEndFlushPolicy(store_config_str=store_config_str), "sync_id", "run_id", "connector")

    for i in range(100000):
        SingletonLogWriter.instance().write("line %d" % i)
        time.sleep(0.1)


if __name__ == "__main__":
    main()


"""
Copyright (c) 2024 valmi.io <https://github.com/valmi-io>

Created Date: Friday, March 17th 2023, 8:50:50 pm
Author: Rajashekar Varkala @ valmi.io

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import copy
import os
import uuid
from pydantic import UUID4
import requests
from requests.adapters import HTTPAdapter, Retry

# TODO: Constants - need to come from current_run_details
HTTP_TIMEOUT = 3  # seconds
MAX_HTTP_RETRIES = 5
CONNECTOR_STRING = "dest"


# desanitise uuid
def du(uuid_str: str) -> UUID4:
    return uuid.UUID(uuid_str.replace("_", "-"))


class ConnectorState:
    def __init__(self, run_time_args={}) -> None:
        # self.num_chunks = run_time_args["chunk_id"]
        # self.begin_records = self.num_chunks * run_time_args["chunk_size"]
        self.records_in_chunk = 0
        self.records_in_this_connector_intance = 0

        self.run_time_args = run_time_args

    def register_chunk(self):
        pass
        # self.num_chunks = self.num_chunks + 1

    def total_records(self):
        pass
        # return self.begin_records + self.records_in_this_connector_intance

    def register_records(self, num_records):
        pass
        # self.records_in_this_connector_intance = num_records
        # self.records_in_chunk = self.total_records() - (self.num_chunks * self.run_time_args["chunk_size"])


class NullEngine:
    def __init__(self) -> None:
        self.connector_state = None
        pass

    def error(self, msg="error"):
        pass

    def success(self):
        pass

    def metric(self):
        pass

    def metric_ext(self, metric_json, chunk_id, commit=False):
        pass

    def current_run_details(self):
        return {}

    def abort_required(self):
        return False


class Engine(NullEngine):
    def __init__(self, *args, **kwargs):
        super(Engine, self).__init__(*args, **kwargs)
        self.engine_url = os.environ["ACTIVATION_ENGINE_URL"]

        self.session_with_retries = requests.Session()
        # retry on all errors except 400 and 401
        status_forcelist = tuple(x for x in requests.status_codes._codes if x > 400 and x not in [400, 401])
        retries = Retry(total=MAX_HTTP_RETRIES, backoff_factor=5, status_forcelist=status_forcelist)
        self.session_with_retries.mount("http://", HTTPAdapter(max_retries=retries))
        self.session_with_retries.mount("https://", HTTPAdapter(max_retries=retries))

        self.session_without_retries = requests.Session()
        status_forcelist = []
        retries = Retry(total=0, backoff_factor=5, status_forcelist=status_forcelist)
        self.session_without_retries.mount("http://", HTTPAdapter(max_retries=retries))
        self.session_without_retries.mount("https://", HTTPAdapter(max_retries=retries))

        run_time_args = self.current_run_details()
        self.connector_state = ConnectorState(run_time_args=run_time_args)

    def current_run_details(self):
        sync_id = du(os.environ.get("DAGSTER_RUN_JOB_NAME", "cf280e5c-1184-4052-b089-f9f41b25138e"))
        # sync_id = du(os.environ.get("DAGSTER_RUN_JOB_NAME", "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"))
        r = self.session_with_retries.get(
            f"{self.engine_url}/syncs/{sync_id}/runs/current_run_details/{CONNECTOR_STRING}",
            timeout=HTTP_TIMEOUT,
        )
        return r.json()

    def metric(self, commit=False):
        pass
        #self.metric_ext({"succeeded": self.connector_state.records_in_chunk}, commit=commit)

    def metric_ext(self, metric_json, chunk_id, commit=False):
        print("Sending metric")
        # TODO: remove the following code. just to make it visible in the UI for now
        metric_copy = copy.deepcopy(metric_json)
        keys_list = list(metric_copy.keys())
        for key in keys_list:
            if key in [
                "upsert",
                "append",
                "create",
                "update",
            ]:
                metric_copy["succeeded"] = metric_copy[key]
                del metric_copy[key]

        payload = {
            "sync_id": self.connector_state.run_time_args["sync_id"],
            "run_id": self.connector_state.run_time_args["run_id"],
            "chunk_id": chunk_id,
            "connector_id": CONNECTOR_STRING,
            "metrics": metric_copy,
            "commit": commit,
        }

        print("payload ", payload)
        if commit:
            r = self.session_with_retries.post(
                f"{self.engine_url}/metrics/",
                timeout=HTTP_TIMEOUT,
                json=payload,
            )
            r.raise_for_status()
        else:
            try:
                r = self.session_without_retries.post(
                    f"{self.engine_url}/metrics/",
                    timeout=HTTP_TIMEOUT,
                    json=payload,
                )
            except Exception:
                pass

    def error(self, msg="error"):
        print("sending error ", msg)
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/status/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json={"status": "failed", "message": msg},
        )
        r.raise_for_status()

    def success(self):
        print("sending success ")
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/status/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json={"status": "success"},
        )
        r.raise_for_status()

    def abort_required(self):
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.get(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/synchronize_connector_engine", timeout=HTTP_TIMEOUT
        )
        return r.json()["abort_required"]

    def checkpoint(self, state):
        print("sending checkpoint")
        sync_id = self.connector_state.run_time_args["sync_id"]
        run_id = self.connector_state.run_time_args["run_id"]
        r = self.session_with_retries.post(
            f"{self.engine_url}/syncs/{sync_id}/runs/{run_id}/state/{CONNECTOR_STRING}/",
            timeout=HTTP_TIMEOUT,
            json=state,
        )
        r.raise_for_status()



from abc import abstractmethod
from collections import defaultdict
import json
from os.path import join
import os
from typing import List
from airbyte_cdk.models import (
    AirbyteMessage
)
from uuid import uuid4
from .constants import MAGIC_DELIM


class SampleWriter:
    writers = {}
    sync_id = None
    store_config_str = None
    run_id = None
    connector = None
    metric_type = None

    @classmethod
    def get_writer_by_metric_type(cls, store_config_str=None,
                                  sync_id=None, run_id=None, connector=None, metric_type=None):
        if sync_id is not None:
            SampleWriter.sync_id = sync_id
        if store_config_str is not None:
            SampleWriter.store_config_str = store_config_str
        if run_id is not None:
            SampleWriter.run_id = run_id
        if connector is not None:
            SampleWriter.connector = connector

        if metric_type is not None and \
            (metric_type not in SampleWriter.writers
             or SampleWriter.writers[metric_type] is None):
            SampleWriter.writers[metric_type] = SampleWriter(SampleWriter.store_config_str,
                                                             ChunkEndFlushPolicy(store_config_str=SampleWriter.store_config_str),
                                                             SampleWriter.sync_id,
                                                             SampleWriter.run_id,
                                                             SampleWriter.connector,
                                                             metric_type)
        if metric_type is not None:
            return SampleWriter.writers[metric_type]
        return None

    def __new__(cls, *args, **kwargs):
        return object.__new__(cls)

    def __init__(self, store_config_str, flush_policy, sync_id, run_id, connector, metric_type):
        self.flush_policy = flush_policy
        self.store_config = json.loads(store_config_str)
        self.records: List[AirbyteMessage] = []
        self.sync_id = sync_id
        self.run_id = run_id
        self.connector = connector
        self.metric_type = metric_type
        self.num_samples_per_code = int(os.environ.get("NUM_SAMPLES_PER_CODE", 10))
        self.sample_counter: dict[str, int] = defaultdict(lambda: 0)

    def reset(self):
        self.records.clear()
        self.flush_policy.reset()

    def check_for_flush(self):
        if self.flush_policy.flush_required():
            self.flush()
            self.reset()

    def write(self, json_log_record):
        rejection_code = json_log_record["record"]["rejection_code"] \
            if "rejected" in json_log_record["record"] and json_log_record["record"]["rejected"] else "200"
        if (self.sample_counter[rejection_code] < self.num_samples_per_code):
            self.sample_counter[rejection_code] += 1
            self.records.append(json_log_record)

    def flush(self):
        if len(self.records) > 0:
            # TODO: do cloud push for aws, s3 when we do k8s
            file_path = join(self.store_config['local']['directory'],
                             self.run_id, "samples", self.connector, f'{self.metric_type}.vals')
            dir_name = os.path.dirname(file_path)
            os.makedirs(dir_name, exist_ok=True)
            with open(file_path, "a+") as f:
                for record in self.records:
                    f.write(record["record"]["rejection_code"] if "rejected" in record["record"] and record["record"]["rejected"] else "200")
                    f.write(MAGIC_DELIM)
                    f.write(record["record"]["synthetic_internal_id"] if "synthetic_internal_id" in record["record"] and record["record"]["synthetic_internal_id"] else str(uuid4()))
                    f.write(MAGIC_DELIM)
                    f.write(json.dumps(record["record"]["data"] if "data" in record["record"] and record["record"]["data"] else {}))
                    f.write(MAGIC_DELIM)
                    f.write(record["record"]["rejection_message"] if "rejection_message" in record["record"] and record["record"]["rejection_message"] else "")
                    f.write(MAGIC_DELIM)
                    f.write(json.dumps(record["record"]["rejection_metadata"] if "rejection_metadata" in record["record"] and record["record"]["rejection_metadata"] else {}))
                    f.write("\n")

    @classmethod
    def data_chunk_flush_callback(cls):
        for k, writer in SampleWriter.writers.items():
            writer.flush_policy.set_data_chunk_flushed()
            writer.check_for_flush()


class FlushPolicy:
    def __init__(self, store_config_str: str):
        self.store_config = json.loads(store_config_str)

    @abstractmethod
    def flush_required(self):
        pass

    @abstractmethod
    def reset(self):
        pass


class ChunkEndFlushPolicy(FlushPolicy):
    def __init__(self, *args, **kwargs):
        super(ChunkEndFlushPolicy, self).__init__(*args, **kwargs)
        self.reset()

    def flush_required(self):
        if self.data_chunk_flushed:
            return True
        return False

    def set_data_chunk_flushed(self):
        self.data_chunk_flushed = True

    def reset(self):
        self.data_chunk_flushed = False


def main():
    from ..valmi_protocol import ValmiFinalisedRecordMessage
    import time
    store_config_str = "{\"provider\": \"local\", \"local\": {\"directory\": \"/tmp/shared_dir/test_logs\", \"max_lines_per_file_hint\" : 100, \"log_flush_interval\": 5}}"

    for i in range(100000):
        sample_writer = SampleWriter.get_writer_by_metric_type(store_config_str, "sync_id", "run_id", "connector", "succeeded")
        sample_writer.write(AirbyteMessage(type=Type.RECORD,
                                           record=ValmiFinalisedRecordMessage(synthetic_internal_id="",
                                            data={"a": "b"},
                                            rejected=False if i%2 == 0 else True,
                                            metric_type="success" if i%2 == 0 else "failed",
                                            rejection_code="200" if i%2 == 0 else "100",
                                            stream="stream",
                                            emitted_at=int(time.time()*1000),
                                            rejection_message="i dont know",
                                            rejection_metadata={"c": "d"},
                                            )))
        time.sleep(0.1)
        sample_writer.data_chunk_flush_callback()


if __name__ == "__main__":
    main()



"""
Copyright (c) 2024 valmi.io <https://github.com/valmi-io>

Created Date: Wednesday, March 8th 2023, 11:38:42 am
Author: Rajashekar Varkala @ valmi.io

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import io
import logging
import sys
from typing import Any, Dict, Iterable, List, Mapping

from airbyte_cdk.destinations import Destination
from airbyte_cdk.models import AirbyteMessage, Type, AirbyteConnectionStatus, Status
from abc import abstractmethod
from valmi_connector_lib.valmi_protocol import ConfiguredValmiDestinationCatalog, ConfiguredValmiCatalog
from airbyte_cdk import AirbyteLogger

logger = logging.getLogger("airbyte")


class ValmiDestination(Destination):
    def parse_args(self, args: List[str]) -> argparse.Namespace:
        parent_parser = argparse.ArgumentParser(add_help=False)
        main_parser = argparse.ArgumentParser()
        subparsers = main_parser.add_subparsers(title="commands", dest="command")

        # spec
        subparsers.add_parser("spec", help="outputs the json configuration specification", parents=[parent_parser])

        # check
        check_parser = subparsers.add_parser(
            "check", help="checks the config can be used to connect", parents=[parent_parser]
        )
        required_check_parser = check_parser.add_argument_group("required named arguments")
        required_check_parser.add_argument(
            "--config", type=str, required=True, help="path to the json configuration file"
        )

        # write
        write_parser = subparsers.add_parser("write", help="Writes data to the destination", parents=[parent_parser])
        # write_parser.add_argument("--state", type=str, required=False, help="path to the json-encoded state file")
        write_required = write_parser.add_argument_group("required named arguments")
        write_required.add_argument("--config", type=str, required=True, help="path to the JSON configuration file")
        write_required.add_argument(
            "--catalog", type=str, required=True, help="path to the configured catalog JSON file"
        )
        write_required.add_argument(
            "--destination_catalog",
            type=str,
            required=True,
            help="path to the configured destination catalog JSON file",
        )
        write_required.add_argument(
            "--state",
            type=str,
            required=False,
            help="path to the state JSON file",
        )

        # discover
        discover_parser = subparsers.add_parser(
            "discover", help="outputs a catalog describing the source's schema", parents=[parent_parser]
        )
        required_discover_parser = discover_parser.add_argument_group("required named arguments")
        required_discover_parser.add_argument(
            "--config", type=str, required=True, help="path to the json configuration file"
        )

        # create object
        create_parser = subparsers.add_parser(
            "create", help="create an object on the destination", parents=[parent_parser]
        )
        required_create_parser = create_parser.add_argument_group("required named arguments")
        required_create_parser.add_argument("--object", type=str, required=True, help="path to the json object file")
        required_create_parser.add_argument(
            "--config", type=str, required=True, help="path to the json configuration file"
        )

        parsed_args = main_parser.parse_args(args)
        cmd = parsed_args.command
        if not cmd:
            raise Exception("No command entered. ")
        elif cmd not in Destination.VALID_CMDS:
            # This is technically dead code since parse_args() would fail if this was the case
            # But it's non-obvious enough to warrant placing it here anyways
            raise Exception(f"Unknown command entered: {cmd}")

        return parsed_args

    # can be overridden to change an input state
    def read_state(self, state_path: str) -> Dict[str, Any]:
        """
        Retrieves the input state of a sync by reading from the specified JSON file.
        """
        if state_path:
            state_obj = self._read_json_file(state_path)
            return state_obj
        return None
 
    def run_cmd(self, parsed_args: argparse.Namespace) -> Iterable[AirbyteMessage]:
        cmd = parsed_args.command

        # if cmd not in ["discover", "write"]:
        if cmd not in ["discover", "write", "create"]:
            for msg in super().run_cmd(parsed_args):
                yield msg
            return

        config = self.read_config(config_path=parsed_args.config)

        if cmd == "discover":
            for msg in self.discover_handler(config, parsed_args):
                yield msg
            return

        if cmd == "create":
            for msg in self.create_handler(config, parsed_args):
                yield msg
            return

        elif cmd == "write":
            state = self.read_state(parsed_args.state)

            # Wrap in UTF-8 to override any other input encodings
            wrapped_stdin = io.TextIOWrapper(sys.stdin.buffer, encoding="utf-8")
            yield from self._run_write(
                # config=config, configured_catalog_path=parsed_args.catalog, input_stream=wrapped_stdin, state=state
                config=config,
                configured_catalog_path=parsed_args.catalog,
                configured_destination_catalog_path=parsed_args.destination_catalog,
                input_stream=wrapped_stdin,
                state=state,
            )
            return

    def create_handler(self, config, parsed_args: argparse.Namespace) -> Iterable[AirbyteMessage]:
        try:
            object_schema = self.read_config(config_path=parsed_args.object)

            create_response=self.create(logger, object_schema)
            
            yield AirbyteMessage(type=Type.CATALOG, create_response=create_response)
        except Exception as err:
            yield AirbyteMessage(
                type=Type.CONNECTION_STATUS,
                connectionStatus=AirbyteConnectionStatus(status=Status.Failed, message=str(err)),
            )

    def discover_handler(self, config, parsed_args: argparse.Namespace) -> Iterable[AirbyteMessage]:
        catalog = self.discover(logger, config)
        yield AirbyteMessage(type=Type.CATALOG, catalog=catalog)

    @abstractmethod
    def write(
        self,
        config: Mapping[str, Any],
        configured_catalog: ConfiguredValmiCatalog,
        input_messages: Iterable[AirbyteMessage],
        configured_destination_catalog: ConfiguredValmiDestinationCatalog,
        logger: AirbyteLogger,
        state: Dict[str, Any] = None,
    ) -> Iterable[AirbyteMessage]:
        """Implement to define how the connector writes data to the destination"""

    def _run_write(
        self,
        config: Mapping[str, Any],
        configured_catalog_path: str,
        configured_destination_catalog_path: str,
        input_stream: io.TextIOWrapper,
        state: Dict[str, Any] = None,
    ) -> Iterable[AirbyteMessage]:
        catalog = ConfiguredValmiCatalog.parse_file(configured_catalog_path)
        destination_catalog = ConfiguredValmiDestinationCatalog.parse_file(configured_destination_catalog_path)
        input_messages = self._parse_input_stream(input_stream)
        logger.info("Begin writing to the destination...")
        yield from self.write(
            # config=config, configured_catalog=catalog, input_messages=input_messages, state=state, logger=logger
            config=config,
            configured_catalog=catalog,
            configured_destination_catalog=destination_catalog,
            input_messages=input_messages,
            logger=logger,
            state=state,
        )
        logger.info("Writing complete.")



from abc import abstractmethod
from collections import defaultdict, namedtuple
from datetime import datetime
from typing import Any, Dict, Iterable, Mapping
from airbyte_cdk import AirbyteLogger
from airbyte_cdk.models import (
    AirbyteMessage,
    AirbyteTraceMessage,
    AirbyteStateMessage,
    TraceType,
    AirbyteErrorTraceMessage,
)
from airbyte_cdk.models.airbyte_protocol import Type, AirbyteStateType

from valmi_connector_lib.valmi_protocol import (
    ConfiguredValmiDestinationCatalog,
    ConfiguredValmiCatalog
)
from valmi_connector_lib.common.run_time_args import RunTimeArgs

HandlerResponseData = namedtuple(
    "HandlerResponseData", ["flushed", "metrics", "emittable_records"], defaults=(False, {}, [])
)


class DestinationWriteWrapper:
    def __init__(
        self,
        logger: AirbyteLogger,
        config: Mapping[str, Any],
        configured_catalog: ConfiguredValmiCatalog,
        configured_destination_catalog: ConfiguredValmiDestinationCatalog,
        state: Dict[str, Any],
    ):
        self.logger = logger
        self.config = config
        self.configured_catalog = configured_catalog
        self.configured_destination_catalog = configured_destination_catalog
        self.run_time_args = RunTimeArgs.parse_obj(config["run_time_args"] if "run_time_args" in config else {})
        self.previous_state = state

    @abstractmethod
    def initialise_message_handling(self) -> None:
        pass

    @abstractmethod
    def handle_message(self, msg: AirbyteMessage, counter: int) -> HandlerResponseData:
        pass

    @abstractmethod
    def finalise_message_handling(self) -> HandlerResponseData:
        pass

    def read_chunk_id_checkpoint(self):
        if self.previous_state is not None \
                and 'state' in self.previous_state \
                and 'data' in self.previous_state['state'] \
                and 'chunk_id' in self.previous_state['state']['data']:
            return self.previous_state['state']['data']['chunk_id'] + 1
        return 1

    def start_message_handling(self, input_messages: Iterable[AirbyteMessage]) -> AirbyteMessage:
        counter: int = 0
        counter_by_type: dict[str, int] = defaultdict(lambda: 0)
        chunk_id = self.read_chunk_id_checkpoint()
        run_time_args = RunTimeArgs.parse_obj(self.config["run_time_args"] if "run_time_args" in self.config else {})

        self.initialise_message_handling()
        for msg in input_messages:
            now = datetime.now()
            if msg.type == Type.RECORD:
                counter = counter + 1

                handler_response = HandlerResponseData()
                try:
                    handler_response = self.handle_message(msg, counter)
                    if handler_response.emittable_records:
                        for record in handler_response.emittable_records:
                            yield AirbyteMessage(
                                type=Type.RECORD,
                                record=record,
                            )
                except Exception as e:
                    yield AirbyteMessage(
                        type=Type.TRACE,
                        trace=AirbyteTraceMessage(
                            type=TraceType.ERROR,
                            error=AirbyteErrorTraceMessage(message=str(e)),
                            emitted_at=int(datetime.now().timestamp()) * 1000,
                        ),
                    )
                    return

                sync_op = msg.record.data["_valmi_meta"]["_valmi_sync_op"]
                if not handler_response.metrics:
                    handler_response = handler_response._replace(metrics={sync_op: 1})

                for op, metric in handler_response.metrics.items():
                    counter_by_type[op] = counter_by_type[op] + metric

                # Commit state only when chunk is finished processing. Flushes are guaranteed for every chunk end, but could be more frequent and even per record for some record
                commit_state = False
                if counter % run_time_args.chunk_size == 0:
                    commit_state = True

                # Aggregate metrics for the current chunk_id and publish
                if counter % run_time_args.records_per_metric == 0 or counter % run_time_args.chunk_size == 0:
                    yield AirbyteMessage(
                        type=Type.STATE,
                        state=AirbyteStateMessage(
                            type=AirbyteStateType.STREAM,
                            data={
                                "records_delivered": counter_by_type,
                                "chunk_id": chunk_id,
                                "finished": False,
                                "commit_state": commit_state,
                                "commit_metric": True,
                            },
                        ),
                    )
                    if counter % run_time_args.chunk_size == 0:
                        counter_by_type.clear()
                        chunk_id = chunk_id + 1

                if (datetime.now() - now).seconds > 5:
                    self.logger.info("A log every 5 seconds - is this required??")

        handler_response = self.finalise_message_handling()
        if handler_response and handler_response.emittable_records:
            for record in handler_response.emittable_records:
                yield AirbyteMessage(
                    type=Type.RECORD,
                    record=record,
                )

        if handler_response:
            for op, metric in handler_response.metrics.items():
                counter_by_type[op] = counter_by_type[op] + metric

        # Sync completed - final state message
        yield AirbyteMessage(
            type=Type.STATE,
            state=AirbyteStateMessage(
                type=AirbyteStateType.STREAM,
                data={
                    "records_delivered": counter_by_type,
                    "chunk_id": chunk_id,
                    "finished": True,
                    "commit_state": True,
                    "commit_metric": True,
                },
            ),
        )
